
import copy

import time
import inspect

import radical.entk  as re
import radical.utils as ru

from   .algorithms import selection_algs, exchange_algs, prepare_algs
from   .algorithms import exchange_alg_prefix
from   .replica    import Replica
from   .utils      import last_task


# ------------------------------------------------------------------------------
#
class Exchange(re.AppManager):

    _t_0    = time.time()
    _glyphs = {re.states.INITIAL:    '+',
               re.states.SCHEDULING: '|',
               re.states.SUSPENDED:  '-',
               re.states.DONE:       ' ',
               re.states.FAILED:     '!',
               re.states.CANCELED:   'X'}


    # --------------------------------------------------------------------------
    #
    def __init__(self, workload, resource, replicas=None):

        self._uid  = ru.generate_id('rx')
        self._prof = ru.Profiler('radical.repex')
        self._prof.prof('create', uid=self._uid)

        self._workload = ru.Config(cfg=workload)
        self._resource = ru.Config(cfg=resource)
        self._replicas = replicas

        # the replicas need to be aware about pre_exec directives
        self._workload.pre_exec = self._resource.pre_exec

        assert(self._workload.config.replicas or self._replicas)
        assert(self._workload.config.cycles)

        self._cycles   = self._workload.config.cycles
        self._waitlist = list()

        if self._replicas:
            self._workload.config.replicas = len(self._replicas)
        else:
            self._replicas = [Replica(workload=self._workload)
                                for _ in range(self._workload.config.replicas)]

        self._pre_alg  = prepare_algs  .get(self._workload.prepare.algorithm)
        self._sel_alg  = selection_algs.get(self._workload.selection.algorithm)
        self._exc_alg  = exchange_algs .get(self._workload.exchange.algorithm)

        # if the configured algorithms are not known (not hard-coded in RX),
        # then assume they point to user specified files and load them
        if not self._pre_alg:
            filename, funcname = self._workload.prepare.algorithm.split(':')
            syms = ru.import_file(filename)
            self._pre_alg = syms['functions'][funcname]

        if not self._sel_alg:
            filename, funcname = self._workload.selection.algorithm.split(':')
            syms = ru.import_file(filename)
            self._sel_alg = syms['functions'][funcname]

        if not self._exc_alg:
            filename, funcname = self._workload.exchange.algorithm.split(':')
            syms = ru.import_file(filename)
            self._exc_alg = syms['functions'][funcname]

        assert(self._pre_alg),  'preparation algorithm missing'
        assert(self._sel_alg),  'selection algorithm missing'
        assert(self._exc_alg),  'exchange algorithm missing'

        rmq_host = str(self._resource.get('rmq_host', 'localhost'))
        rmq_port = int(self._resource.get('rmq_port', '5672'))
        rmq_user = str(self._resource.get('rmq_user','guest'))
        rmq_pass = str(self._resource.get('rmq_pass','guest'))
        re.AppManager.__init__(self, autoterminate=True,
                                     hostname=rmq_host, port=rmq_port,
                                     username=rmq_user, password=rmq_pass)

        for r in self._replicas:
            r._initialize(check_ex=self._check_exchange,
                          check_res=self._check_resume,
                          sid=self.sid, prof=self._prof)

        self._lock = ru.Lock(name='rx')

        rd = copy.deepcopy(self._resource)
        if 'rmq_host' in rd: del(rd['rmq_host'])
        if 'rmq_port' in rd: del(rd['rmq_port'])
        if 'pre_exec' in rd: del(rd['pre_exec'])

        self.resource_desc = rd

        self._log  = ru.Logger('radical.repex')
        self._dout = open('dump.log', 'a')
        self._dump(msg='startup')

        # run the replica pipelines
        self.workflow = set(self._replicas)


    # --------------------------------------------------------------------------
    #
    def run(self):
        '''
        run the replica exchange pipelines, and after all is done, fetch the
        requested output data
        '''

        # run the preparator, set resulting data as `shared_data`, and begin to
        # work
        fnames = ru.as_list(self._pre_alg(self._workload))

        if self._workload.data.inputs not in fnames:
            fnames.append(self._workload.data.inputs)

        # write exchange algorithm to disk (once), and then stage with every
        # exchange task
        self._ex_alg_file = 'exchange_algorithm.py'
        with open('%s/%s' % (self._workload.data.inputs, self._ex_alg_file),
                  'w') as fout:
            fout.write(exchange_alg_prefix % (inspect.getsource(self._exc_alg),
                                              self._exc_alg.__name__))

        self.shared_data = fnames
        re.AppManager.run(self)


    # --------------------------------------------------------------------------
    #
    def _dump(self, msg=None, special=None, glyph=None ):

        if not self._dout:
            return

        if not msg:
            msg = ''

        if not self._dout:
            return

        self._dout.write(' | %7.2f |' % (time.time() - self._t_0))
        for r in self._replicas:
            if special and r in special:
                self._dout.write('%s' % glyph)
            else:
                self._dout.write('%s' % self._glyphs[r.state])
        self._dout.write('| %s\n' % msg)
        self._dout.flush()


    # --------------------------------------------------------------------------
    #
    def terminate(self):

        self._log.debug('exc term')
        self._dump(msg='terminate', special=self._replicas, glyph='=')
        if self._dout:
            self._dout.close()
            self._dout = None

        # we are done!
        re.AppManager.terminate(self)


    # --------------------------------------------------------------------------
    #
    def _check_exchange(self, replica):

        # method races when concurrently triggered by multpiple replicas
        with self._lock:

            self._waitlist.append(replica)

            ex_list   = None
            new_wlist = None

            # invoke the user defined selection algorithm
            try:
                ex_list, new_wlist = self._sel_alg(
                        waitlist=self._waitlist,
                        criteria=self._workload.selection,
                        replica=replica)
                self._log.debug('sel: %4d -> %4d + %4d = %4d',
                        len(self._waitlist), len(ex_list), len(new_wlist),
                        len(ex_list) + len(new_wlist))
            except Exception as e:
                self._log.exception('selection algorithm failed: %s' % e)

            # check if the user found something to exchange
            if not ex_list:
                # nothing to do, suspend this replica and wait until we get more
                # candidates and can try again
                self._log.debug('%5s %s no  - suspend',
                                replica.rid, replica._uid)
                replica.suspend()
                self._dump()

                # waiting for more replicas only makes sense if any others are
                # still in `SCHEDULING` state
                states = dict()
                for r in self._replicas:
                    if r.state not in states: states[r.state] = 1
                    else                    : states[r.state] += 1
                self._log.debug('=== %s', ['%s=%s' % (k, v) for k, v in states.items()])

                if states.get(re.states.SCHEDULING):
                    # some more replicas are active - wait for those to complete
                    # the exchange list
                    return

                # did not find any active replics, thus the exchange list will
                # never grow and a new exchange will never happen - terminate
                self._log.warn('=== terminating due to lack of active replicas')
                raise RuntimeError('terminating due to lack of active replicas')

            # Seems we got an exchange list - check it: exchange list and
            # new wait list must be proper partitions of the original waitlist:
            #   - make sure no replica is lost
            #   - make sure that replicas are not in both lists
            assert(new_wlist is not None)
            missing = len(self._waitlist) - len(ex_list) - len(new_wlist)
            if missing:
                raise ValueError('%d replicas went missing' % missing)

            for r in self._waitlist:
                if r not in ex_list and r not in new_wlist:
                    raise ValueError('replica %s (%s) missing'
                                    % r, r.properties)

            if replica not in ex_list:
                raise ValueError('active replica (%s) not in exchange list %s)'
                                % (replica.rid, [r.rid for r in ex_list]))

            # lists are valid - use them
            self._waitlist = new_wlist

            self._log.debug('%5s %s yes - exchange', replica.rid, replica._uid)
            msg = " > %s: %s" % (replica.rid, [r.rid for r in ex_list])
            self._dump(msg=msg, special=ex_list, glyph='v')

            # we have a set of exchange candidates.  The current replica is
            # tasked to host the exchange task.
            replica.add_ex_stage(ex_list, self._ex_alg_file, self.sid)


    # --------------------------------------------------------------------------
    #
    def _check_resume(self, replica):

        self._dump()
        self._log.debug('%5s %s check resume', replica.rid, replica._uid)

        resumed = list()  # list of resumed replica IDs

        msg = " < %s: %s" % (replica.rid,
                             [r.rid for r in replica.exchange_list])
        self._dump(msg=msg, special=replica.exchange_list, glyph='^')

        exchange = last_task(replica)

        # after a successfull exchange we revive all participating replicas.
        # For those replicas which did not yet reach min cycles, add an md
        # stage, all others we let die and add a new md stage for them.
        for _replica in replica.exchange_list:

            if _replica.cycle <= self._cycles:
                last = bool(_replica.cycle == self._cycles)
                _replica.add_md_stage(exchanged_from=exchange,
                                      sid=self.sid, last=last)

            # Make sure we don't resume the current replica
            if replica.rid != _replica.rid:

                self._log.debug('%5s %s resume', _replica.rid, replica._uid)
                if _replica.state == re.states.SUSPENDED:
                    _replica.resume()
                resumed.append(_replica.uid)

        return resumed


# ------------------------------------------------------------------------------





import radical.utils as ru


# ------------------------------------------------------------------------------
#
def expand_ln(to_link, src_sbox, tgt_sbox, rid, cycle, task_id=None):

    expand = {'rid'  : rid,
              'cycle': cycle}

    if not src_sbox: src_sbox = '.'
    if not tgt_sbox: tgt_sbox = '.'

    ret = list()
    for data in ru.as_list(to_link):
        src, tgt = data.split('>')
        try:
            src = src.strip() % expand
            tgt = tgt.strip() % expand
        except Exception as e:
            raise RuntimeError('expansion error: %s : %s : %s'
                              % (src, tgt, expand))
        # if task_id is None:
        #     ret.append('%s/%s > %s/%s_%s'
        #               % (src_sbox, src, tgt_sbox, tgt, task_id))
        # else:
        #
        ret.append('%s/%s > %s/%s' % (src_sbox, src, tgt_sbox, tgt))

    return ret


# ------------------------------------------------------------------------------
#
def last_task(replica):

    cs = replica.current_stage

    if cs >= len(replica.stages):
        cs -= 1

    assert(cs < len(replica.stages))
    tasks = replica.stages[cs].tasks

    assert(tasks)
    assert(len(tasks) == 1)

    return list(tasks)[0]


# ------------------------------------------------------------------------------





import copy

import radical.utils as ru
import radical.entk  as re

from .utils import expand_ln, last_task


# ------------------------------------------------------------------------------
#
class Replica(re.Pipeline):
    '''
    A `Replica` is an EnTK pipeline which consists of alternating md and
    exchange stages.  The initial setup is for one MD stage - Exchange and more
    MD stages get added depending on runtime conditions.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, workload, properties=None):

        self._workload  = workload
        self._check_ex  = None
        self._check_res = None

        if not properties:
            properties  = dict()

        self._rid       = ru.generate_id('rep.%(counter)04d', ru.ID_CUSTOM)

        # this is inefficient at scale...
        self._prof      = ru.Profiler('radical.entk')
        self._prof.prof('create', uid=self._rid)

        self._props     = properties
        self._cycle     = -1    # increased when adding md stage
        self._ex_list   = None  # list of replicas used in exchange step

        re.Pipeline.__init__(self)
        self.name = 'p.%s' % self.rid
        self._log = ru.Logger('radical.repex')


    # --------------------------------------------------------------------------
    #
    def _initialize(self, check_ex, check_res, sid, prof):
        '''
        This method should only be called by the Exchange class upon
        initialization.
        '''

        self._check_ex  = check_ex
        self._check_res = check_res

        # add an initial md stage
        self.add_md_stage(sid=sid)

        self._prof.close()  # switch to the global profiler
        self._prof = prof

        self._prof.prof('init', uid=self.rid)


    # --------------------------------------------------------------------------
    #
    @property
    def rid(self):
        return self._rid

    @property
    def cycle(self):
        return self._cycle

    @property
    def properties(self):
        return self._props


    # --------------------------------------------------------------------------
    #
    @property
    def exchange_list(self):
        return self._ex_list


    # --------------------------------------------------------------------------
    #
    def add_md_stage(self, exchanged_from=None, sid=None, last=False):

        self._prof.prof('add_md_start', uid=self.rid)
        self._cycle += 1
        self._log.debug('%5s %s add md', self.rid, self._uid)

      # task = re.Task(from_dict=self._workload['md'])
      # task.name = 'mdtsk-%s-%s' % (self.rid, self.cycle)
        sandbox     = '%s.%04d.md' % (self.rid, self.cycle)
        link_inputs = list()

        # link initial data
        link_inputs += expand_ln(self._workload.md.inputs,
                     'pilot:///%s' % self._workload.data.inputs,
                     'task://', self.rid, self.cycle)

        if self._cycle == 0:
            # link initial data
            link_inputs += expand_ln(self._workload.md.inputs_0,
                         'pilot:///%s' % self._workload.data.inputs,
                         'task://',
                         self.rid, self.cycle)
        else:
            # get data from previous task
            t = last_task(self)
            if exchanged_from:
                self._log.debug('Exchange from %s', exchanged_from.name)
                link_inputs += expand_ln(self._workload.md.ex_2_md,
                        'pilot:///%s' % (exchanged_from.sandbox),
                        'task://',
                        self.rid, self.cycle)
            else:
                # FIXME: this apparently can't happen
                link_inputs += expand_ln(self._workload.md.md_2_md,
                         'resource:///%s' % (t.sandbox),
                         'task://',
                         self.rid, self.cycle)

        copy_outputs = expand_ln(self._workload.md.outputs,
                         'task://',
                         'client:///%s' % self._workload.data.outputs,
                         self.rid, self.cycle)

        if last:
            copy_outputs += expand_ln(self._workload.md.outputs_n,
                         'task://',
                         'client:///%s' % self._workload.data.outputs,
                         self.rid, self.cycle)

        # TODO: filter out custom keys from that dict before deepcopy
        env   = {'REPEX_RID'   : str(self.rid),
                 'REPEX_CYCLE' : str(self.cycle)}
        tds   = copy.deepcopy(self._workload['md']['descriptions'])
        first = 0
        last  = len(tds) - 1
        for idx, td in enumerate(tds):

            stage = re.Stage()
            task  = re.Task()
            td    = ru.expand_env(td, env=env)

            for k,v in td.items():
                setattr(task, k, v)

            if self._workload.pre_exec:
                if task.pre_exec:
                    task.pre_exec.extend(self._workload.pre_exec)
                else:
                    task.pre_exec.extend = self._workload.pre_exec

            task.name    = '%s.%04d.%02d.md' % (self.rid, self.cycle, idx)
            task.sandbox = sandbox

            if idx == first:
                task.link_input_data = link_inputs

            if idx == last:
                task.download_output_data = copy_outputs
                stage.post_exec = self.check_exchange

            stage.add_tasks(task)
            self.add_stages(stage)
            self._log.debug('%5s add md: %s', self.rid, task.name)

        self._prof.prof('add_md_stop', uid=self.rid)


    # --------------------------------------------------------------------------
    #
    def check_exchange(self):
        '''
        after an md cycle, record its completion and check for exchange
        '''

        self._prof.prof('chk_ex_start', uid=self.rid)
        self._log.debug('%5s check_exchange %s', self.rid, self._uid)
        self._check_ex(self)
        self._prof.prof('chk_ex_stop', uid=self.rid)


    # --------------------------------------------------------------------------
    #
    def add_ex_stage(self, exchange_list, ex_alg, sid):

        self._prof.prof('add_ex_start', uid=self.rid)
        self._log.debug('%5s add ex: %s', self.rid,
                        [r.rid for r in exchange_list])

        self._ex_list = exchange_list

        task = re.Task()
        task.executable = 'python3'
        task.arguments  = [ex_alg, '-r', self.rid, '-c', self.cycle] \
                        + ['-e'] + [r.rid for r in exchange_list] \
                        + ['-d'] + [d for d in self._workload.exchange.ex_data]

        if self._workload.pre_exec:
            task.pre_exec = self._workload.pre_exec

        # link alg
        link_inputs = ['pilot:///%s/%s' % (self._workload.data.inputs, ex_alg)]

        # link exchange data
        for r in exchange_list:

            t = last_task(r)
            self._log.debug('Exchage: %s, Task Name: %s Sandbox %s', r.name,
                             t.name, t.sandbox)
            link_inputs += expand_ln(self._workload.exchange.md_2_ex,
                                     # FIXME: how to get absolute task sbox?
                                     #        rep.0000.0000:/// ...
                                     #        i.e., use task ID as schema
                                     'pilot:///%s' % t.sandbox,
                                     'task://', r.rid, r.cycle)

        task.link_input_data = link_inputs

        task.name    = '%s.%04d.ex' % (self.rid, self.cycle)
        task.sandbox = '%s.%04d.ex' % (self.rid, self.cycle)

        self._log.debug('%5s added ex: %s, input data: %s',
                        self.rid, task.name, task.link_input_data)

        stage = re.Stage()
        stage.add_tasks(task)
        stage.post_exec = self.check_resume

        self.add_stages(stage)
        self._prof.prof('add_ex_stop', uid=self.rid)


    # --------------------------------------------------------------------------
    #
    def check_resume(self):
        '''
        after an ex cycle, trigger replica resumption
        '''

        self._prof.prof('chk_res_start', uid=self.rid)
        self._log.debug('%5s check_resume %s', self.rid, self._uid)
        ret = self._check_res(self)
        self._log.debug('%5s check_resume %s: %s', self.rid, self._uid, ret)

        self._prof.prof('chk_res_stop', uid=self.rid, msg=ret)
        return ret


# ------------------------------------------------------------------------------






# ------------------------------------------------------------------------------
#
from .replica    import Replica
from .exchange   import Exchange
from .algorithms import *


# ------------------------------------------------------------------------------
#
import os
import radical.utils as ru

pwd  = os.path.dirname (__file__)
root = "%s" % pwd
version_short, version_detail, version_base, \
        version_branch, sdist_name, sdist_path = ru.get_version(paths=[root])
version = version_short

logger = ru.Logger('radical.repex')
logger.info('radical.repex        version: %s' % version_detail)

# ------------------------------------------------------------------------------





import radical.utils as ru

SELECT_1D       = '1D'
SELECT_TEST     = 'TEST'
EXCHANGE_RANDOM = 'RANDOM'

_log = ru.Logger('radical.repex')


# ------------------------------------------------------------------------------
#
def select_replicas_1D(waitlist, criteria, replica):
    '''
    replica selection algorithm: out of the list of eligible replicas, select
    those which should be part of an exchange step.

    Arguments:

      - waitlist: a list of replica objects which are eligible for exchange
      - criteria: dictionary of selection criteria to control the algorthm's
    '''

    try:
        _log.debug('criteria: %s' % criteria)
        _log.debug('waitlist: %s' % [r.rid for r in waitlist])

        # get required parameters
        ex_size = criteria['exchange_size']

        # check if size of wait list suffices
        if len(waitlist) < ex_size:

            # not enough replicas to attempt exchange
            return [], waitlist
        elif len(waitlist) > ex_size:
            # waitlist somehow grew too large

            raise ValueError('The waitlist is larger than maximum permissible'
                             'exchange list size')

        # we have enough replicas!  Remove all as echange candidates from the
        # waitlist and return them!
        exchange_list = list()
        for r in waitlist:
            exchange_list.append(r)

        # empty the waitlist to start collecting new candidates
        new_waitlist = list()

        _log.debug('exchange: %s' % [r.rid for r in exchange_list])
        _log.debug('new wait: %s' % [r.rid for r in new_waitlist])

        return exchange_list, new_waitlist


    except Exception:

        _log.exception('replica selection failed')

        # on failure, return the unchanged waitlist and an empty selection
        return [], waitlist


# ------------------------------------------------------------------------------
#
def select_replicas_test(waitlist, criteria, replica):

    if len(waitlist) < criteria['exchange_size']:
        return [], waitlist

    return [r for r in waitlist], []


# ------------------------------------------------------------------------------
#
def exchange_by_random():
    '''
    This method is run as workload of exchange tasks.  It will receive two
    arguments: the number of replicas to exchange, and the cycle (?).
    '''

    import sys
    import random

    replicas = int(sys.argv[1])
    cycle    = int(sys.argv[2])

    exchange_list_1 = list(range(replicas))
    exchange_list_2 = list(range(replicas))

    random.shuffle(exchange_list_1)
    random.shuffle(exchange_list_2)

    exchangePairs = zip(exchange_list_2, exchange_list_2)

    with open('exchangePairs_%d.dat' % cycle, 'w') as f:
        for p in exchangePairs:
            line = ' '.join(str(x) for x in p)
            f.write(line + '\n')


# ------------------------------------------------------------------------------
#
exchange_alg_prefix = '''
#!/usr/bin/env python

import sys


# ------------------------------------------------------------------------------
#
# exchange algorithm
#
%s


# ------------------------------------------------------------------------------
#
# exchange algorithm call context
#
if __name__ == '__main__':

    rid     = None
    cycle   = None
    ex_list = list()
    ex_data = list()

    arg_mode = None
    for arg in sys.argv[1:]:
        if   arg == '-r': arg_mode = 'rid'
        elif arg == '-c': arg_mode = 'cycle'
        elif arg == '-e': arg_mode = 'ex_list'
        elif arg == '-d': arg_mode = 'ex_data'
        else:
            if   arg_mode == 'rid'    : rid   = arg
            elif arg_mode == 'cycle'  : cycle = int(arg)
            elif arg_mode == 'ex_list': ex_list.append(arg)
            elif arg_mode == 'ex_data': ex_data.append(arg)

    # call the exchange algorithm inserted above
    %s(rid, cycle, ex_list, ex_data)


# ------------------------------------------------------------------------------

'''


# ------------------------------------------------------------------------------
#
selection_algs = {
        SELECT_1D   : select_replicas_1D,
        SELECT_TEST : select_replicas_test
}

exchange_algs = {
        EXCHANGE_RANDOM : exchange_by_random
}

prepare_algs = {
}


# ------------------------------------------------------------------------------
